//: DYNARRAY.CPP -- The advantages of an array, with the
// dynamic flexibility of a linked list.
//. Modified to use templates.
//. From "Using C++, 2nd edition" by Bruce Eckel (Osborne/McGraw-Hill 1992)
#pragma option -Jgd
#include "dynarray.h"
#include <iostream.h>
#include <stdlib.h>  // exit()
#pragma option -Jgd  // to generate public definitions

#include "ostring.h"
// typedef DynArray<String> dummy_type_string_vector;  // generate definition
typedef DynArray<String, 10> dummy_type_string_vector;  // generate definition
// default argument not working here....

template<class T, int chunk>
void DynArray<T, chunk>::error(char * msg) {
  cerr << "DynArray error: " << msg << endl;
  exit(1);
}

template<class T, int chunk>
DynArray<T, chunk>::DynArray() {
  size = chunk;
  cursor = 0;
  array = new T*[size];  // allocate a chunk of pointers
  memset(array, 0, sizeof(T*) * size); // zero the memory
}

template<class T, int chunk>
DynArray<T, chunk>::~DynArray() {
  for(int i = 0; i < size; i++)
    delete array[i];
  delete array;
}

template<class T, int chunk>
int DynArray<T, chunk>::add(T * new_element) {
  // Put it in the first empty space avaialable:
  for(int i = 0; i < size; i++)
    if(array[i] == 0) {
      array[i] = new_element;
      return i;
    }
  // at this point, no space was found.  Add new space:
  int tempsize = size + chunk;  // increase space by chunk
  T ** temp = new T*[tempsize];
  memset(temp, 0, sizeof(T*) * tempsize); // zero the memory
  // copy the old array over:
  for(i = 0; i < size; i++)
    temp[i] = array[i];
  temp[i = size] = new_element;  // put at start of empty space
  delete array;  // free old memory
  array = temp;
  size = tempsize;
  return i;
}

template<class T, int chunk>
int DynArray<T, chunk>::remove(T * rp) {
  for(int i = 0; i < size; i++)
    if( array[i] == rp ) {
      delete array[i];  // must have been made on the heap!
      array[i] = 0;
      return 1;
    }
  return 0; // not found
}

template<class T, int chunk>
int DynArray<T, chunk>::remove(int ri) {
  if( ri < 0 || ri >= size)
    error("remove index out of range");
  // check to see if there's an element at that slot:
  if( array[ri] ) {
    delete array[ri];  // must have been made on the heap!
    array[ri] = 0;
    return 1;
  }
  return 0; // not found
}

template<class T, int chunk>
void DynArray<T, chunk>::reset() {
  cursor = 0;
  while( (array[cursor] == 0) && (cursor < size - 1) )
    cursor++;  // find the first nonzero element
}

template<class T, int chunk>
T * DynArray<T, chunk>::next() {
  if( cursor == size -1 ) // last element, no next.
    return 0;
  // Not at the end. Increment until you find a
  // non-empty slot, or the end:
  while(array[++cursor] == 0)
    if ( cursor == size -1 )  // no more elements in list
      return 0;
  return array[cursor];
}

template<class T, int chunk>
T * DynArray<T, chunk>::operator[](int x) {
  if(x < 0 || x >= size)
    error("operator[] -- index out of range");
  return array[x];  // even if it's empty...
}

template<class T, int chunk>
int DynArray<T, chunk>::count() {
  int cnt = 0;
  for(int x=0; x < size; x++)
    if(array[x])
      cnt++;
  return cnt;
}
